class DeprecationError(Exception):
    """An exception that indicates an error when using deprecated params"""


class NexusflowAIError(Exception):
    """An exception that indicates an error which creating the NexusflowAI client"""
